/*
   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2005 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.0 of the PHP license,       |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_0.txt.                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Omar Kilani <omar@tinysofa.com>                              |
   +----------------------------------------------------------------------+
   $Id: php_js.c,v 1.2 2005/09/29 10:49:46 omar Exp $
*/

#ifdef COMPILE_DL_JS
#define HAVE_JS 1
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_JS

#ifdef ZEND_WIN32
#  define _WINSOCK2API_           /* using winsock.h instead of winsock2.h */
#  define READDIR_H
#  define _INET_H_
#  define _NETDB_H_
#endif

#include <jsapi.h>

#include "php.h"
#include "php_ini.h"
#include "zend_extensions.h"
#include "ext/standard/info.h"
#include "SAPI.h"
#include "php_js.h"

ZEND_BEGIN_MODULE_GLOBALS(js)
  JSRuntime      *js;
  JSContext      *cx;
  JSObject       *global;
ZEND_END_MODULE_GLOBALS(js)

PHP_FUNCTION(js_eval);

#ifdef ZTS
#define JSG(v) TSRMG(js_globals_id, zend_js_globals *, v)
#else
#define JSG(v) (js_globals.v)
#endif

ZEND_DECLARE_MODULE_GLOBALS(js);

JSClass _php_js_globalClass = {
    "Global", 0,
    JS_PropertyStub,  JS_PropertyStub,JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,  JS_FinalizeStub
};

static JSBool
_php_js_system_print(JSContext *cx, JSObject *obj, int argc, jsval *argv, jsval* rval){
    int i;
    JSString *str;

    for(i=0;i<argc;i++){
        str = JS_ValueToString(cx, argv[i]);
        if(!str) return JS_FALSE;
        zend_printf("%s%s", i ? " ":"", JS_GetStringBytes(str));
    }

    return JS_TRUE;
}

static JSBool
_php_js_system_gc(JSContext *cx, JSObject *obj, int argc, jsval *argv, jsval* rval){
    JS_GC(cx);
    return JS_TRUE;
}

JSFunctionSpec _php_js_globalMethods[] = {
    {"print",   _php_js_system_print,   0, 0},
    {"gc",      _php_js_system_gc,      0, 0},
    { NULL,     NULL,           0, 0}
};

static JSBool
_php_js_evaluate_string(JSContext *cx, JSObject *obj, char *string, size_t len, jsval *rval)
{
    JS_EvaluateScript(cx, obj, string, len, "php_js", 0, rval);
}

static void
_php_js_errorReporter(JSContext *cx, const char *message, JSErrorReport *report){
    if(report == NULL)
        zend_error(E_WARNING, "[php_js] %s\n", message);
    else
        zend_error(E_WARNING, "[php_js] %s: line %i: error: %s\n",
            report->filename, report->lineno, message);
}


/* Creates JavaScript interpreter if it was not created before */
static JSRuntime* php_js_init(TSRMLS_D)
{
  JSRuntime* my_js = JSG(js);
  JSContext* my_cx = JSG(cx);
  JSObject * my_global = JSG(global);

  if (my_js == NULL) {
    /* initialize the JS run time, and return result in my_js */
    my_js = JS_NewRuntime(8L * 1024L * 1024L);

    /* if my_js does not have a value, return null */
    if (my_js == NULL)
        return NULL;
  }

  if (my_cx == NULL)
  {
    /* create a context and associate it with the JS run time */
    my_cx = JS_NewContext(my_js, 8192);

    /* if my_cx does not have a value, return null */
    if (my_cx == NULL)
        return NULL;

    JS_SetErrorReporter(my_cx, _php_js_errorReporter);     /* error reporter */

    my_global = JS_NewObject(my_cx, &_php_js_globalClass, NULL, NULL);

    /* initialize the built-in JS objects and the global object */
    JS_InitStandardClasses(my_cx, my_global);

    JS_DefineFunctions(my_cx, my_global, _php_js_globalMethods);

    JSG(js) = my_js;
    JSG(cx) = my_cx;
    JSG(global) = my_global;
  }
  return my_js;
}

/* Destroys JavaScript interpriter if it was created brfore */
static void php_js_destroy(TSRMLS_D)
{
  JSRuntime* my_js = JSG(js);
  JSContext* my_cx = JSG(cx);

  if (my_js != NULL) {
    JS_GC(my_cx);
    JS_DestroyContext(my_cx);
    JS_DestroyRuntime(my_js);
    JSG(cx) = NULL;
    JSG(js) = NULL;
  }
}

static zval* php_js_jsval_to_zval_r(JSContext      *my_cx,
                                       jsval       *val,
                                       zend_bool    assoc) 
{
    zval *return_value;

    MAKE_STD_ZVAL(return_value);

    if (JSVAL_IS_NULL(*val))
    {
        ZVAL_NULL(return_value);
    }
    else if (JSVAL_IS_BOOLEAN(*val))
    {
        ZVAL_BOOL(return_value, JSVAL_TO_BOOLEAN(*val));
    }
    else if (JSVAL_IS_DOUBLE(*val))
    {
        ZVAL_DOUBLE(return_value, *(JSVAL_TO_DOUBLE(*val)));
    }
    else if (JSVAL_IS_INT(*val))
    {
        ZVAL_LONG(return_value, JSVAL_TO_INT(*val));
    }
    else if (JSVAL_IS_OBJECT(*val))
    {
      JSObject *obj = JSVAL_TO_OBJECT(*val);
      if (JS_IsArrayObject(my_cx, obj) == JS_FALSE)
      {
        JSIdArray *properties = JS_Enumerate(my_cx, obj);
        zval *mval;
        jsval aval, vval;
        jsint i = 0, flags = 0;

        if (assoc) {
          array_init(return_value);
        } else {
          object_init(return_value);
        }

        for (i = 0; i < properties->length; i++)
        {
          JSString *str;
          JS_IdToValue(my_cx, properties->vector[i], &aval);
          if (aval) {
            JS_CheckAccess(my_cx, obj, properties->vector[i], 0, &vval, &flags);
            mval = php_js_jsval_to_zval_r(my_cx, &vval, assoc);
            str = JS_ValueToString(my_cx, aval);
          } else {
            MAKE_STD_ZVAL(mval);
            ZVAL_NULL(mval);
          }

          if (assoc) {
            add_assoc_zval(return_value, JS_GetStringBytes(str), mval);
          } else {
            add_property_zval(return_value, JS_GetStringBytes(str), mval);
#if PHP_MAJOR_VERSION >= 5
            ZVAL_DELREF(mval);
#endif
          }
        }
        JS_DestroyIdArray(my_cx, properties);
      }
      else
      {
        zval *mval;
        jsval aval;
        int i = 0, l;

        array_init(return_value);
        JS_GetArrayLength(my_cx, obj, &l);
        for (i = 0; i < l; i++) {
          JS_GetElement(my_cx, obj, i, &aval);
          if (aval) {
            mval = php_js_jsval_to_zval_r(my_cx, &aval, assoc);
          } else {
            MAKE_STD_ZVAL(mval);
            ZVAL_NULL(mval);
          }
          add_index_zval(return_value, i, mval);
        }
      }
    }
    else if (JSVAL_IS_STRING(*val))
    {
        JSString *str = JSVAL_TO_STRING(*val);
        char *s =  JS_GetStringBytes(str);
        ZVAL_STRING(return_value, s, 1);
    }
    else
    {
        ZVAL_NULL(return_value);
    }

    return return_value;
}

/****************************************************************************/

static void php_js_init_globals(zend_js_globals *js_globals)
{
  js_globals->js = NULL;
  js_globals->cx = NULL;
  js_globals->global = NULL;
}

PHP_MINIT_FUNCTION(js)
{
  ZEND_INIT_MODULE_GLOBALS(js, php_js_init_globals, NULL);
}

PHP_MSHUTDOWN_FUNCTION(js)
{
  return SUCCESS;
}

PHP_RSHUTDOWN_FUNCTION(js)
{
  php_js_destroy(TSRMLS_C);
  return SUCCESS;
}

PHP_MINFO_FUNCTION(js)
{
  JSRuntime* my_js = php_js_init(TSRMLS_C);

  php_info_print_table_start();
  php_info_print_table_header(2, "JavaScript support", "enabled");
  php_info_print_table_row(2, "Version", PHP_JS_VERSION);
  php_info_print_table_end();
}

function_entry js_functions[] = {
    PHP_FE(js_eval, NULL)
    {NULL, NULL, NULL}  /* Must be the last line in json_functions[] */
};

/* eval($javascript_code)
   Evaluates JavaScript code and returns the result.
   Produces a error if code has JavaScript errors */
ZEND_FUNCTION(js_eval)
{
  char* js_code;
  int   js_code_len;
  zend_bool assoc = 0;
  jsval rval;
  JSString *str;
  zval *z;

  if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|b",
                            &js_code, &js_code_len, &assoc) != FAILURE) {
    JSRuntime* my_js = php_js_init(TSRMLS_C);
    JSContext* my_cx = JSG(cx);
    JSObject * my_global = JSG(global);
    _php_js_evaluate_string(my_cx, my_global, js_code, js_code_len, &rval);
    z = php_js_jsval_to_zval_r(my_cx, &rval, assoc);
    if (!z) {
        RETURN_NULL();
    }

    *return_value = *z;

    FREE_ZVAL(z);
  }
  else
  {
    RETURN_NULL();
  }
}

zend_module_entry js_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
  STANDARD_MODULE_HEADER,
#endif
  "js",
  js_functions,
  PHP_MINIT(js),
  PHP_MSHUTDOWN(js),
  NULL,
  PHP_RSHUTDOWN(js),
  PHP_MINFO(js),
#if ZEND_MODULE_API_NO >= 20010901
  PHP_JS_VERSION,
#endif
  STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_JS
ZEND_GET_MODULE(js)
#endif

#endif
